#!/bin/sh

# Regression testing script for CTeX
# Step through a list of files
#  Compile, run, and check the output of each expected-to-work test
#  Compile and check the error of each expected-to-fail test

# Path to the LLVM interpreter
LLI="lli"
#LLI="/usr/local/opt/llvm/bin/lli"

# Path to the LLVM compiler
LLC="llc"

# Path to the C compiler
CC="cc"

# Path to the ctex compiler.  Usually "../src/ctex.native"
CTEX="./src/ctex.native"

LIBPATH="./src/binom.o ./src/gcd.o"

LATEX_COV="/bin/sh ./scripts/ctex2tex.sh"

PDFLATEX="pdflatex"

# Set time limit for all operations
ulimit -t 30

globallog=run_ctex.log
rm -f $globallog

error=0
globalerror=0

keep=0
pdf=1
build=1
run=1

Usage() {
    echo "Usage: run_ctex.sh [options] [.ctex files]"
    echo "-b    Only build executable file from .ctex file"
    echo "-r    Only build and run executable file from .ctex file"
    echo "-k    Keep intermediate files"
    echo "-p    Only generate pdf file from .ctex file"
    echo "-h    Print this help"
    exit 1
}

SignalError() {
    if [ $error -eq 0 ] ; then
	echo "FAILED"
	error=1
    fi
    echo "  $1"
}

# Run <args>
# Report the command, run it, and report any errors
Run() {
    echo $* 1>&2
    eval $* || {
	SignalError "$1 failed on $*"
	return 1
    }
}

Check() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.ctex//'`

    echo "###### Working on $basename"

    echo 1>&2
    echo "###### Working on $basename" 1>&2

    generatedfiles=""

    if [ $pdf -eq 1 ] ; then
        generatedfiles="$generatedfiles ${basename}.tex ${basename}.aux ${basename}.out ${basename}.log ${basename}.synctex.gz" &&
        Run "$LATEX_COV" "$1" > "${basename}.tex" &&
        Run "$PDFLATEX" "-output-directory=./" "${basename}.tex"
    fi

    if [ $build -eq 1 ] ; then
        generatedfiles="$generatedfiles ${basename}.ll ${basename}.s" &&
        Run "$CTEX" < "$1" ">" "${basename}.ll" &&
        Run "$LLC" "-relocation-model=pic" "${basename}.ll" ">" "${basename}.s" &&
        Run "$CC" "-o" "${basename}.exe" "${basename}.s" "${LIBPATH}" "-lm"
    fi
    
    if [ $run -eq 1 ] ; then
        generatedfiles="$generatedfiles ${basename}.exe" &&
        Run "./${basename}.exe"
    fi
    
    # Report the status and clean up the generated files

    if [ $error -eq 0 ] ; then
	if [ $keep -eq 0 ] ; then
	    rm -f $generatedfiles
	fi
    echo "###### FINISHED SUCCESSFULLY"
	echo "###### FINISHED SUCCESSFULLY" 1>&2
    else
    echo "###### FAILED"
	echo "###### FAILED" 1>&2
	globalerror=$error
    fi
}

while getopts kbrph c; do
    case $c in
	k) # Keep intermediate files
	    keep=1
	    ;;
    b) 
	    run=0
        pdf=0
	    ;;
    r) 
        pdf=0
	    ;;
    p) 
	    build=0
        run=0
	    ;;
	h) # Help
	    Usage
	    ;;
    esac
done

shift `expr $OPTIND - 1`

LLIFail() {
  echo "Could not find the LLVM interpreter \"$LLI\"."
  echo "Check your LLVM installation and/or modify the LLI variable in testall.sh"
  exit 1
}

which "$LLI" >> $globallog || LLIFail

for file in $@
do
    Check $file 2>> $globallog
done

exit $globalerror
