type paren_op = Abs | Floor | Ceil

type log_like_op =
  | Lg
  | Ln
  | Sqrt
  | Sin
  | Cos
  | Tan
  | Arcsin
  | Arccos
  | Arctan
  | Sinh
  | Cosh
  | Tanh
  | Cot
  | Sec
  | Csc
  | Coth

type unop = Uplus | Uminus

type mult_like_op = Mult | Div | Mod

type add_like_op = Plus | Minus

type ident =
  | SId of string
  | SSId of string * int

type func = FId of ident | Gcd | Min | Max

type frac_like_op = Frac | Binom

type large_op = Sum | Prod

type comp_op = Lt | Gt | Leq | Geq | Eq | Neq | Nmid | Mid

type logic_binop = And | Or

type short_atom = ShortLiteral of int

and long_atom =
  | IntLiteral of int
  | FloatLiteral of float
  | ParenthLike of paren_op * expr_calc

and atom = ShortAtom of short_atom | LongAtom of long_atom

and expr_pow =
  | Atom of atom
  | PowerFuncAtom of atom * expr_calc
  | PowerFuncId of expr_calc * expr_calc

and expr_log =
  | ExprPow of expr_pow
  | LogLike of log_like_op * expr_log
  | LogLikeC of log_like_op * expr_calc
  | Log of expr_calc * expr_log
  | LogC of expr_calc * expr_calc

and expr_impl_mult =
  | ExprLog of expr_log
  | ImplMult of expr_impl_mult * expr_log
  | ImplMultC of expr_calc * expr_log
  | ImplMultMC of expr_impl_mult * expr_calc
  | ImplMultCMC of expr_calc * expr_calc
  | SpecImplNoId of ident * expr_calc

and expr_spec = ident * ident

and expr_unary =
  | ExprImplMult of expr_impl_mult
  | Id of ident
  | ExprNoId of expr_no_id
  | Unop of unop * expr_unary
  | UnopC of unop * expr_calc

and expr_mult =
  | ExprUnary of expr_unary
  | MultLike of expr_mult * mult_like_op * expr_unary
  | MultLikeMC of expr_mult * mult_like_op * expr_calc
  | MultLikeC of expr_calc * mult_like_op * expr_unary
  | MultLikeCMC of expr_calc * mult_like_op * expr_calc

and expr_add =
  | ExprMult of expr_mult
  | AddLike of expr_add * add_like_op * expr_mult
  | AddLikeMC of expr_add * add_like_op * expr_calc
  | AddLikeC of expr_calc * add_like_op * expr_mult
  | AddLikeCMC of expr_calc * add_like_op * expr_calc

and expr_no_id =
  | ExprAdd of expr_add
  | Call of func * expr_calc list
  | FracLike of frac_like_op * expr_calc * expr_calc
  (* | LargeOp of large_op * ident * expr_calc * expr_calc * expr_calc *)

and expr_calc =
  | Id of ident
  | SpecImplId of expr_spec
  | ExprNoId of expr_no_id

and expr_comp = Comp of expr_calc * comp_op * expr_calc

and expr_logic_atom =
  | ExprComp of expr_comp
  | ParenLogic of expr_logic
  | Not of expr_logic_atom

and expr_logic =
  | LogicAtom of expr_logic_atom
  | Logic of expr_logic * logic_binop * expr_logic_atom

type sutie = CaseSutie of stmt * expr_logic

and stmt =
  | Expr of expr_calc
  | Assign of ident * expr_calc
  | Print of expr_calc
  | FuncDef of ident * ident list * stmt
  | Case of sutie list
  | StmtClosure of stmt list

type program = stmt list

(* Pretty-printing functions *)

let string_of_log_like_op = function
  | Lg -> "\\lg "
  | Ln -> "\\ln "
  | Sqrt -> "\\sqrt "
  | Sin -> "\\sin "
  | Cos -> "\\cos "
  | Tan -> "\\tan "
  | Arcsin -> "\\arcsin "
  | Arccos -> "\\arccos "
  | Arctan -> "\\arctan "
  | Sinh -> "\\sinh "
  | Cosh -> "\\cosh "
  | Tanh -> "\\tanh "
  | Cot -> "\\cot "
  | Sec -> "\\sec "
  | Csc -> "\\csc "
  | Coth -> "\\coth "

let string_of_uop = function Uplus -> "+" | Uminus -> "-"

let string_of_multi_like_op = function
  | Mult -> "*"
  | Div -> "/"
  | Mod -> "\\mod "

let string_of_add_like_op = function Plus -> "+" | Minus -> "-"

let string_of_frac_like_op = function
  | Frac -> "\\frac "
  | Binom -> "\\binom "

let string_of_ident = function
 | SId ss -> ss
 | SSId (ss, sub) -> ss ^ (string_of_int sub)

let string_of_func = function
  | FId s -> string_of_ident s
  | Gcd -> "\\gcd "
  | Min -> "\\min "
  | Max -> "\\max "

let string_of_comp_op = function
  | Lt -> "< "
  | Gt -> "> "
  | Leq -> "\\leq "
  | Geq -> "\\geq "
  | Eq -> "= "
  | Neq -> "\\neq "
  | Nmid -> "\\nmid "
  | Mid -> "\\mid "

let string_of_logic_op = function And -> "\\wedge " | Or -> "\\vee "

let rec string_of_short_atom = function ShortLiteral l -> string_of_int l

and string_of_long_atom = function
  | IntLiteral l -> string_of_int l
  | FloatLiteral l -> string_of_float l
  | ParenthLike (op, e) -> (
      match op with
      | Abs -> "|" ^ string_of_expr_calc e ^ "|"
      | Floor -> "\\lfloor " ^ string_of_expr_calc e ^ " \\rfloor"
      | Ceil -> "\\lceil " ^ string_of_expr_calc e ^ " \\rceil")

and string_of_atom = function
  | ShortAtom a -> string_of_short_atom a
  | LongAtom a -> string_of_long_atom a

and string_of_expr_pow = function
  | Atom a -> string_of_atom a
  | PowerFuncAtom (a, e) -> string_of_atom a ^ "^" ^ string_of_expr_calc e
  | PowerFuncId (a, e) -> string_of_expr_calc a ^ "^" ^ string_of_expr_calc e

and string_of_expr_log = function
  | ExprPow e -> string_of_expr_pow e
  | LogLike (op, e) ->
      string_of_log_like_op op ^ "(" ^ string_of_expr_log e ^ ")"
  | LogLikeC (op, e) ->
      string_of_log_like_op op ^ "(" ^ string_of_expr_calc e ^ ")"
  | Log (a, e) ->
      "\\log_" ^ string_of_expr_calc a ^ " " ^ string_of_expr_log e
  | LogC (a, e) ->
      "\\log_" ^ string_of_expr_calc a ^ " " ^ string_of_expr_calc e

and string_of_expr_impl_mult = function
  | ExprLog e -> string_of_expr_log e
  | ImplMult (e1, e2) ->
      string_of_expr_impl_mult e1 ^ "•" ^ string_of_expr_log e2
  | ImplMultC (e1, e2) ->
      string_of_expr_calc e1 ^ "•" ^ string_of_expr_log e2
  | ImplMultMC (e1, e2) ->
      string_of_expr_impl_mult e1 ^ "•" ^ string_of_expr_calc e2
  | ImplMultCMC (e1, e2) ->
      string_of_expr_calc e1 ^ "•" ^ string_of_expr_calc e2
  | SpecImplNoId (s, e) -> string_of_ident s ^ "(" ^ string_of_expr_calc e ^ ")"

and string_of_expr_spec (s1, s2) = string_of_ident s1 ^ "(" ^ string_of_ident s2 ^ ")"

and string_of_expr_unary = function
  | ExprImplMult e -> string_of_expr_impl_mult e
  | Id s -> string_of_ident s
  | ExprNoId e -> string_of_expr_no_id e
  | Unop (op, e) -> string_of_uop op ^ string_of_expr_unary e
  | UnopC (op, e) -> string_of_uop op ^ string_of_expr_calc e

and string_of_expr_mult = function
  | ExprUnary e -> string_of_expr_unary e
  | MultLike (e1, op, e2) ->
      string_of_expr_mult e1
      ^ string_of_multi_like_op op
      ^ string_of_expr_unary e2
  | MultLikeMC (e1, op, e2) ->
      string_of_expr_mult e1
      ^ string_of_multi_like_op op
      ^ string_of_expr_calc e2
  | MultLikeC (e1, op, e2) ->
      string_of_expr_calc e1
      ^ string_of_multi_like_op op
      ^ string_of_expr_unary e2
  | MultLikeCMC (e1, op, e2) ->
      string_of_expr_calc e1
      ^ string_of_multi_like_op op
      ^ string_of_expr_calc e2

and string_of_expr_add = function
  | ExprMult e -> string_of_expr_mult e
  | AddLike (e1, op, e2) ->
      string_of_expr_add e1 ^ string_of_add_like_op op
      ^ string_of_expr_mult e2
  | AddLikeMC (e1, op, e2) ->
      string_of_expr_add e1 ^ string_of_add_like_op op
      ^ string_of_expr_calc e2
  | AddLikeC (e1, op, e2) ->
      string_of_expr_calc e1 ^ string_of_add_like_op op
      ^ string_of_expr_mult e2
  | AddLikeCMC (e1, op, e2) ->
      string_of_expr_calc e1 ^ string_of_add_like_op op
      ^ string_of_expr_calc e2

and string_of_expr_no_id = function
  | ExprAdd e -> string_of_expr_add e
  | Call (f, el) ->
      string_of_func f ^ "("
      ^ String.concat ", " (List.map string_of_expr_calc el)
      ^ ")"
  | FracLike (op, e1, e2) ->
      string_of_frac_like_op op ^ "{" ^ string_of_expr_calc e1 ^ "} {"
      ^ string_of_expr_calc e2 ^ "}"
  (* | LargeOp (op, s, es, ee, e) -> "Unimpl" *)

and string_of_expr_calc = function
  | Id s -> string_of_ident s
  | SpecImplId s -> string_of_expr_spec s
  | ExprNoId e -> string_of_expr_no_id e

and string_of_expr_comp = function
  | Comp (e1, op, e2) ->
      string_of_expr_calc e1 ^ string_of_comp_op op ^ string_of_expr_calc e2

and string_of_expr_logic_atom = function
  | ExprComp e -> string_of_expr_comp e
  | ParenLogic e -> "(" ^ string_of_expr_logic e ^ ")"
  | Not e -> "\\neg " ^ string_of_expr_logic_atom e

and string_of_expr_logic = function
  | LogicAtom e -> string_of_expr_logic_atom e
  | Logic (e1, op, e2) ->
      string_of_expr_logic e1 ^ string_of_logic_op op
      ^ string_of_expr_logic_atom e2

let rec string_of_suite = function
  | CaseSutie (s, e) ->
      string_of_stmt s ^ "&" ^ string_of_expr_logic e ^ "\\\\"

and string_of_stmt = function
  | Expr e -> string_of_expr_calc e ^ "\\\\"
  | Assign (s, e) -> string_of_ident s ^ "=" ^ string_of_expr_calc e ^ "\\\\"
  | Print e -> "print " ^ string_of_expr_calc e ^ "\n"
  | FuncDef (s, fl, ss) ->
      string_of_ident s ^ "(" ^ String.concat ", " (List.map string_of_ident fl) ^ ") = " ^ string_of_stmt ss
  | Case sl ->
      "\\begin{cases}\n"
      ^ String.concat "\n" (List.map string_of_suite sl)
      ^ "\n\\end{cases}\n"
  | StmtClosure sl ->
      "\\begin{split}\n"
      ^ String.concat "\n" (List.map string_of_stmt sl)
      ^ "\n\\end{split}\n"

let string_of_program sl = String.concat "\n" (List.map string_of_stmt sl)
