open Ast

type typ = Int | Bool | Float

type bind = typ * string

type s_atom_x =
  | SId of string
  | SIntLiteral of int
  | SFloatLiteral of float
  | SParen of s_expr_calc
  | SParenthLike of paren_op * s_expr_calc

and s_atom = typ * s_atom_x

and s_expr_pow_x = SAtom of s_atom | SPowerFunc of s_atom * s_expr_calc

and s_expr_pow = typ * s_expr_pow_x

and s_expr_log_x =
  | SExprPow of s_expr_pow
  | SLogLike of log_like_op * s_expr_log
  | SLog of s_expr_calc * s_expr_log

and s_expr_log = typ * s_expr_log_x

and s_expr_impl_mult_x =
  | SExprLog of s_expr_log
  | SImplMult of s_expr_impl_mult * s_expr_log
  | SCall of func * s_expr_calc list

and s_expr_impl_mult = typ * s_expr_impl_mult_x

and s_expr_unary_x =
  | SExprImplMult of s_expr_impl_mult
  | SUnop of unop * s_expr_unary
  | SId of string
  | SParen of s_expr_calc

and s_expr_unary = typ * s_expr_unary_x

and s_expr_mult_x =
  | SExprUnary of s_expr_unary
  | SMultLike of s_expr_mult * mult_like_op * s_expr_unary

and s_expr_mult = typ * s_expr_mult_x

and s_expr_add_x =
  | SExprMult of s_expr_mult
  | SAddLike of s_expr_add * add_like_op * s_expr_mult

and s_expr_add = typ * s_expr_add_x

and s_expr_calc_x =
  | SExprAdd of s_expr_add
  | SCall of func * s_expr_calc list
  | SCallLargeOp of string
  | SFracLike of frac_like_op * s_expr_calc * s_expr_calc
(*  | SLargeOp of large_op * bind * s_expr_calc * s_expr_calc * s_expr_calc*)

and s_expr_calc = typ * s_expr_calc_x

and s_expr_comp_x = SComp of s_expr_calc * comp_op * s_expr_calc

and s_expr_comp = typ * s_expr_comp_x

and s_expr_logic_atom_x =
  | SExprComp of s_expr_comp
  | SParenLogic of s_expr_logic
  | SNot of s_expr_logic_atom

and s_expr_logic_atom = typ * s_expr_logic_atom_x

and s_expr_logic_x =
  | SLogicAtom of s_expr_logic_atom
  | SLogic of s_expr_logic * logic_binop * s_expr_logic_atom

and s_expr_logic = typ * s_expr_logic_x

type sfunc_decl = {
  styp : typ;
  sfname : string;
  sformals : bind list;
  sframe : bind list;
  slocals : bind list;
  sbody : sstmt;
}

and ssutie_x = SCaseSutie of sstmt * s_expr_logic

and ssutie = typ * ssutie_x

and sstmt_x =
  | SExpr of s_expr_calc
  | SAssign of string * s_expr_calc
  | SPrint of s_expr_calc
  | SFuncDef of sfunc_decl
  | SCase of ssutie list
  | SStmtClosure of sstmt list

and sstmt = Some of typ * sstmt_x | None

type sprogram = sfunc_decl list

let string_of_typ = function
  | Int -> "int"
  | Bool -> "bool"
  | Float -> "float"

let string_of_bind (t, s) = string_of_typ t ^ ":" ^ s

let rec string_of_s_atom (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SId s -> s
    | SIntLiteral l -> string_of_int l
    | SFloatLiteral l -> string_of_float l
    | SParen e -> "(" ^ string_of_s_expr_calc e ^ ")"
    | SParenthLike (op, e) -> (
        match op with
        | Abs -> "|" ^ string_of_s_expr_calc e ^ "|"
        | Floor -> "\\lfloor " ^ string_of_s_expr_calc e ^ " \\rfloor"
        | Ceil -> "\\lceil " ^ string_of_s_expr_calc e ^ " \\rceil"))
  ^ ")"

and string_of_s_expr_pow (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SAtom a -> string_of_s_atom a
    | SPowerFunc (a, e) -> string_of_s_atom a ^ "^" ^ string_of_s_expr_calc e)
  ^ ")"

and string_of_s_expr_log (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SExprPow e -> string_of_s_expr_pow e
    | SLogLike (op, e) ->
        string_of_log_like_op op ^ "(" ^ string_of_s_expr_log e ^ ")"
    | SLog (a, e) ->
        "\\log_" ^ string_of_s_expr_calc a ^ " " ^ string_of_s_expr_log e)
  ^ ")"

and string_of_s_expr_impl_mult (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SExprLog e -> string_of_s_expr_log e
    | SImplMult (e1, e2) ->
        string_of_s_expr_impl_mult e1 ^ "•" ^ string_of_s_expr_log e2
    | SCall (f, el) ->
        string_of_func f ^ "("
        ^ String.concat ", " (List.map string_of_s_expr_calc el)
        ^ ")")
  ^ ")"

and string_of_s_expr_unary (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SExprImplMult e -> string_of_s_expr_impl_mult e
    | SUnop (op, e) -> string_of_uop op ^ string_of_s_expr_unary e
    | SId s -> s
    | SParen e -> "(" ^ string_of_s_expr_calc e ^ ")")
  ^ ")"

and string_of_s_expr_mult (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SExprUnary e -> string_of_s_expr_unary e
    | SMultLike (e1, op, e2) ->
        string_of_s_expr_mult e1
        ^ string_of_multi_like_op op
        ^ string_of_s_expr_unary e2)
  ^ ")"

and string_of_s_expr_add (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SExprMult e -> string_of_s_expr_mult e
    | SAddLike (e1, op, e2) ->
        string_of_s_expr_add e1 ^ string_of_add_like_op op
        ^ string_of_s_expr_mult e2)
  ^ ")"

and string_of_s_expr_calc (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SExprAdd e -> string_of_s_expr_add e
    | SCall (f, el) ->
        string_of_func f ^ "("
        ^ String.concat ", " (List.map string_of_s_expr_calc el)
        ^ ")"
    | SFracLike (op, e1, e2) ->
        string_of_frac_like_op op ^ "{" ^ string_of_s_expr_calc e1 ^ "} {"
        ^ string_of_s_expr_calc e2 ^ "}"
    )
  ^ ")"

and string_of_s_expr_comp (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SComp (e1, op, e2) ->
        string_of_s_expr_calc e1 ^ string_of_comp_op op
        ^ string_of_s_expr_calc e2)
  ^ ")"

and string_of_s_expr_logic_atom (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SExprComp e -> string_of_s_expr_comp e
    | SParenLogic e -> "(" ^ string_of_s_expr_logic e ^ ")"
    | SNot e -> "\\neg " ^ string_of_s_expr_logic_atom e)
  ^ ")"

and string_of_s_expr_logic (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SLogicAtom e -> string_of_s_expr_logic_atom e
    | SLogic (e1, op, e2) ->
        string_of_s_expr_logic e1 ^ string_of_logic_op op
        ^ string_of_s_expr_logic_atom e2)
  ^ ")"

let rec string_of_sfunc_decl fdecl =
  string_of_typ fdecl.styp ^ " " ^ fdecl.sfname ^ "("
  ^ String.concat ", " (List.map snd fdecl.sformals)
  ^ ")\n<Frame: "
  ^ String.concat "; " (List.map string_of_bind fdecl.sframe)
  ^ ">\n[Locals:"
  ^ String.concat "; " (List.map string_of_bind fdecl.slocals)
  ^ "]\n{\n"
  ^ string_of_sstmt fdecl.sbody
  ^ "}\n"

and string_of_ssutie (t, s) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match s with
    | SCaseSutie (s, e) ->
        string_of_sstmt s ^ " when " ^ string_of_s_expr_logic e)
  ^ ")"

and string_of_sstmt = function
  | Some (t, s) ->
      "(" ^ string_of_typ t ^ " : "
      ^ (match s with
        | SExpr e -> string_of_s_expr_calc e ^ "\\\\"
        | SAssign (s, e) -> s ^ "=" ^ string_of_s_expr_calc e ^ "\\\\"
        | SPrint e -> "print " ^ string_of_s_expr_calc e ^ "\n"
        | SFuncDef fd -> "define " ^ string_of_sfunc_decl fd ^ "\n"
        | SCase sl ->
            "\\begin{cases}\n"
            ^ String.concat "\n" (List.map string_of_ssutie sl)
            ^ "\n\\end{cases}\n"
        | SStmtClosure sl ->
            "\\begin{split}\n"
            ^ String.concat "\n" (List.map string_of_sstmt sl)
            ^ "\n\\end{split}\n")
      ^ ")"
  | None -> raise (Failure "Empty body found in function")

let string_of_sprogram sl =
  String.concat "\n" (List.map string_of_sfunc_decl sl)
