open Ast
open Sast
module StringMap = Map.Make (String)

let check_program prog =
  let type_of_identifier lb s =
    try StringMap.find s lb
    with Not_found -> raise (Failure ("undeclared identifier " ^ s))
  in

  let find_func lf s =
    try StringMap.find s lf
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  let atom_of_expr_calc e : s_atom = (fst e, SParen e)
  and power_of_atom e = (fst e, SAtom e)
  and log_of_power e = (fst e, SExprPow e)
  and imp_of_log e = (fst e, SExprLog e)
  and unary_of_imp e = (fst e, SExprImplMult e)
  and mult_of_unary e = (fst e, SExprUnary e)
  and add_of_mult e = (fst e, SExprMult e)
  and calc_of_add e = (fst e, SExprAdd e) in
  let impl t1 t2 e1' e2' : s_expr_impl_mult =
    match (t1, t2) with
    | Int, Int -> (Int, SImplMult (e1', e2'))
    | Float, Float | Int, Float | Float, Int -> (Float, SImplMult (e1', e2'))
    | _ -> raise (Failure "Invalid bool type appears")
  in
  let mult op t1 t2 e1' e2' =
    match op with
    | Mod -> (Int, SMultLike (e1', Mod, e2'))
    | (Mult | Div) when t1 = t2 && t1 = Int -> (Int, SMultLike (e1', op, e2'))
    | (Mult | Div) when t1 != Bool && t2 != Bool ->
        (Float, SMultLike (e1', op, e2'))
    | _ -> raise (Failure "Invalid bool type appears")
  in
  let add op t1 t2 e1' e2' =
    match op with
    | (Plus | Minus) when t1 = t2 && t1 = Int ->
        (Int, SAddLike (e1', op, e2'))
    | (Plus | Minus) when t1 != Bool && t2 != Bool ->
        (Float, SAddLike (e1', op, e2'))
    | _ -> raise (Failure "Invalid bool type appears")
  in
  let rec check_short_atom lb lf e =
    match e with ShortLiteral l -> (Int, SIntLiteral l)
  and check_long_atom lb lf e =
    match e with
    | IntLiteral l -> (Int, SIntLiteral l)
    | FloatLiteral fl -> (Float, SFloatLiteral fl)
    | ParenthLike (op, e) -> (
        match op with
        | Abs ->
            let ((t, se) as e') = check_expr_calc lb lf e in
            (t, SParenthLike (Abs, e'))
        | Floor | Ceil ->
            let ((t, se) as e') = check_expr_calc lb lf e in
            (Int, SParenthLike (op, e')))
  and check_atom lb lf e =
    match e with
    | ShortAtom a -> check_short_atom lb lf a
    | LongAtom a -> check_long_atom lb lf a
  and check_expr_pow lb lf e =
    match e with
    | Atom a ->
        let ((t, sa) as a') = check_atom lb lf a in
        (t, SAtom a')
    | PowerFuncAtom (a1, a2) ->
        let ((t1, sa1) as a1') = check_atom lb lf a1
        and ((t2, sa2) as a2') = check_expr_calc lb lf a2 in
        (Float, SPowerFunc (a1', a2'))
    | PowerFuncId (a1, a2) ->
        let ((t1, sa1) as a1') = check_expr_calc lb lf a1
        and ((t2, sa2) as a2') = check_expr_calc lb lf a2 in
        (Float, SPowerFunc ((t1, SParen a1'), a2'))
  and check_expr_log lb lf e =
    match e with
    | ExprPow e ->
        let ((t, se) as e') = check_expr_pow lb lf e in
        (t, SExprPow e')
    | LogLike (op, e) -> (Float, SLogLike (op, check_expr_log lb lf e))
    | LogLikeC (op, e) ->
        ( Float,
          SLogLike
            ( op,
              log_of_power
                (power_of_atom (atom_of_expr_calc (check_expr_calc lb lf e)))
            ) )
    | Log (ec, e) ->
        (Float, SLog (check_expr_calc lb lf ec, check_expr_log lb lf e))
    | LogC (ec, e) ->
        ( Float,
          SLog
            ( check_expr_calc lb lf ec,
              log_of_power
                (power_of_atom (atom_of_expr_calc (check_expr_calc lb lf e)))
            ) )
  and check_expr_impl_mult lb lf e =
    match e with
    | ExprLog e ->
        let ((t, se) as e') = check_expr_log lb lf e in
        (t, SExprLog e')
    | ImplMult (e1, e2) ->
        let ((t1, se1) as e1') = check_expr_impl_mult lb lf e1
        and ((t2, se2) as e2') = check_expr_log lb lf e2 in
        impl t1 t2 e1' e2'
    | ImplMultC (e1, e2) ->
        let ((t1, se1) as e1') = check_expr_calc lb lf e1
        and ((t2, se2) as e2') = check_expr_log lb lf e2 in
        let e1' =
          imp_of_log (log_of_power (power_of_atom (atom_of_expr_calc e1')))
        in
        impl t1 t2 e1' e2'
    | ImplMultMC (e1, e2) ->
        let ((t1, se1) as e1') = check_expr_impl_mult lb lf e1
        and ((t2, se2) as e2') = check_expr_calc lb lf e2 in
        let e2' = log_of_power (power_of_atom (atom_of_expr_calc e2')) in
        impl t1 t2 e1' e2'
    | ImplMultCMC (e1, e2) ->
        let ((t1, se1) as e1') = check_expr_calc lb lf e1
        and ((t2, se2) as e2') = check_expr_calc lb lf e2 in
        let e1' =
          imp_of_log (log_of_power (power_of_atom (atom_of_expr_calc e1')))
        and e2' = log_of_power (power_of_atom (atom_of_expr_calc e2')) in
        impl t1 t2 e1' e2'
    | SpecImplNoId (s, e) -> (
        match StringMap.find_opt (string_of_ident s) lb with
        | Some t ->
            check_expr_impl_mult lb lf (ImplMultCMC (Id s, e))
        | None -> (
            match StringMap.find_opt (string_of_ident s) lf with
            | Some fd ->
                let param_length = List.length fd.sformals in
                if 1 != param_length then
                  raise
                    (Failure
                       ("expecting "
                       ^ string_of_int param_length
                       ^ " arguments in "))
                else (fd.styp, SCall (FId s, [ check_expr_calc lb lf e ]))
            | None -> raise (Failure ("Unknown bind:" ^ (string_of_ident s)))))
  and check_expr_unary lb lf e =
    match e with
    | ExprImplMult e ->
        let ((t, se) as e') = check_expr_impl_mult lb lf e in
        (t, SExprImplMult e')
    | Id s ->
        let t = type_of_identifier lb (string_of_ident s) in
        (t, SId (string_of_ident s))
    | ExprNoId e ->
        let ((t, se) as e') = check_expr_no_id lb lf e in
        (t, SParen e')
    | Unop (op, e) ->
        let ((t, se) as e') = check_expr_unary lb lf e in
        (t, SUnop (op, e'))
    | UnopC (op, e) ->
        let ((t, se) as e') = check_expr_calc lb lf e in
        ( t,
          SUnop
            ( op,
              unary_of_imp
                (imp_of_log
                   (log_of_power (power_of_atom (atom_of_expr_calc e')))) )
        )
  and check_expr_mult lb lf e =
    match e with
    | ExprUnary e ->
        let ((t, se) as e') = check_expr_unary lb lf e in
        (t, SExprUnary e')
    | MultLike (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_mult lb lf e1
        and ((t2, se2) as e2') = check_expr_unary lb lf e2 in
        mult op t1 t2 e1' e2'
    | MultLikeMC (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_mult lb lf e1
        and ((t2, se2) as e2') = check_expr_calc lb lf e2 in
        let e2' =
          unary_of_imp
            (imp_of_log
               (log_of_power (power_of_atom (atom_of_expr_calc e2'))))
        in
        mult op t1 t2 e1' e2'
    | MultLikeC (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_calc lb lf e1
        and ((t2, se2) as e2') = check_expr_unary lb lf e2 in
        let e1' =
          mult_of_unary
            (unary_of_imp
               (imp_of_log
                  (log_of_power (power_of_atom (atom_of_expr_calc e1')))))
        in
        mult op t1 t2 e1' e2'
    | MultLikeCMC (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_calc lb lf e1
        and ((t2, se2) as e2') = check_expr_calc lb lf e2 in
        let e1' =
          mult_of_unary
            (unary_of_imp
               (imp_of_log
                  (log_of_power (power_of_atom (atom_of_expr_calc e1')))))
        and e2' =
          unary_of_imp
            (imp_of_log
               (log_of_power (power_of_atom (atom_of_expr_calc e2'))))
        in
        mult op t1 t2 e1' e2'
  and check_expr_add lb lf e =
    match e with
    | ExprMult e ->
        let ((t, se) as e') = check_expr_mult lb lf e in
        (t, SExprMult e')
    | AddLike (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_add lb lf e1
        and ((t2, se2) as e2') = check_expr_mult lb lf e2 in
        add op t1 t2 e1' e2'
    | AddLikeMC (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_add lb lf e1
        and ((t2, se2) as e2') = check_expr_calc lb lf e2 in
        let e2' =
          mult_of_unary
            (unary_of_imp
               (imp_of_log
                  (log_of_power (power_of_atom (atom_of_expr_calc e2')))))
        in
        add op t1 t2 e1' e2'
    | AddLikeC (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_calc lb lf e1
        and ((t2, se2) as e2') = check_expr_mult lb lf e2 in
        let e1' =
          add_of_mult
            (mult_of_unary
               (unary_of_imp
                  (imp_of_log
                     (log_of_power (power_of_atom (atom_of_expr_calc e1'))))))
        in
        add op t1 t2 e1' e2'
    | AddLikeCMC (e1, op, e2) ->
        let ((t1, se1) as e1') = check_expr_calc lb lf e1
        and ((t2, se2) as e2') = check_expr_calc lb lf e2 in
        let e1' =
          add_of_mult
            (mult_of_unary
               (unary_of_imp
                  (imp_of_log
                     (log_of_power (power_of_atom (atom_of_expr_calc e1'))))))
        and e2' =
          mult_of_unary
            (unary_of_imp
               (imp_of_log
                  (log_of_power (power_of_atom (atom_of_expr_calc e2')))))
        in
        add op t1 t2 e1' e2'
  and check_expr_no_id lb lf e =
    match e with
    | ExprAdd e ->
        let ((t, se) as e') = check_expr_add lb lf e in
        (t, SExprAdd e')
    | Call (fname, args) -> (
        match fname with
        | FId (s : ident) ->
            let fd = find_func lf (string_of_ident s) in
            let param_length = List.length fd.sformals in
            if List.length args != param_length then
              raise
                (Failure
                   ("expecting "
                   ^ string_of_int param_length
                   ^ " arguments in "))
            else
              (fd.styp, SCall (FId s, List.map (check_expr_calc lb lf) args))
        | Gcd ->
            if List.length args != 2 then
              raise (Failure "expecting 2 arguments for gcd")
            else (Int, SCall (Gcd, List.map (check_expr_calc lb lf) args))
        | Min | Max ->
            if List.length args < 2 then
              raise (Failure "expecting >2 arguments for min/max")
            else if
              List.exists
                (fun e -> fst (check_expr_calc lb lf e) = Float)
                args
            then (Float, SCall (fname, List.map (check_expr_calc lb lf) args))
            else (Int, SCall (fname, List.map (check_expr_calc lb lf) args)))
    | FracLike (op, ac1, ac2) -> (
        let t1, sa1 = check_expr_calc lb lf ac1
        and t2, sa2 = check_expr_calc lb lf ac2 in
        match op with
        | Frac -> (Float, SFracLike (Frac, (t1, sa1), (t2, sa2)))
        | Binom -> (Int, SFracLike (Binom, (t1, sa1), (t2, sa2))))
(*    | LargeOp (op, s, se, ee, ce) ->*)
(*          (Int, SIntLiteral 123)*)
(*        (Int, SCallLargeOp ":random")*)
(*        ( Float,*)
(*          SLargeOp*)
(*            ( op,*)
(*              (Int, s),*)
(*              check_expr_calc lb lf se,*)
(*              check_expr_calc lb lf ee,*)
(*              check_expr_calc lb lf ce ) )*)
  and check_expr_calc lb lf e =
    match e with
    | Id s ->
        let t = type_of_identifier lb (string_of_ident s) in
        let e = (t, SId (string_of_ident s)) in
        calc_of_add (add_of_mult (mult_of_unary e))
    | SpecImplId e -> (
        let s1, s2 = e in
        match StringMap.find_opt (string_of_ident s1) lb with
        | Some t ->
            calc_of_add
              (add_of_mult
                 (mult_of_unary
                    (unary_of_imp
                       (check_expr_impl_mult lb lf
                          (ImplMultCMC (Id s1, Id s2))))))
        | None -> (
            match StringMap.find_opt (string_of_ident s1) lf with
            | Some fd ->
                let param_length = List.length fd.sformals in
                if 1 != param_length then
                  raise
                    (Failure
                       ("expecting "
                       ^ string_of_int param_length
                       ^ " arguments in "))
                else
                  let t = type_of_identifier lb (string_of_ident s2) in
                  let e = (t, SId (string_of_ident s2)) in
                  ( fd.styp,
                    SCall
                      ( FId s1,
                        [ calc_of_add (add_of_mult (mult_of_unary e)) ] ) )
            | None -> raise (Failure ("Unknown bind:" ^ (string_of_ident s1)))))
    | ExprNoId e -> check_expr_no_id lb lf e
  and check_expr_comp lb lf e =
    match e with
    | Comp (e1, op, e2) ->
        (Bool, SComp (check_expr_calc lb lf e1, op, check_expr_calc lb lf e2))
  and check_expr_logic_atom lb lf e =
    match e with
    | ExprComp e ->
        let ((t, se) as e') = check_expr_comp lb lf e in
        (t, SExprComp e')
    | ParenLogic e ->
        let ((t, se) as e') = check_expr_logic lb lf e in
        (t, SParenLogic e')
    | Not e -> (Bool, SNot (check_expr_logic_atom lb lf e))
  and check_expr_logic lb lf e =
    match e with
    | LogicAtom e ->
        let ((t, se) as e') = check_expr_logic_atom lb lf e in
        (t, SLogicAtom e')
    | Logic (e1, op, e2) ->
        ( Bool,
          SLogic
            (check_expr_logic lb lf e1, op, check_expr_logic_atom lb lf e2)
        )
  in

  let add_bind s t lb lf =
    if StringMap.find_opt s lf != None then
      raise (Failure (s ^ "has been bound to a function previously"))
    else
      match StringMap.find_opt s lb with
      | Some nt ->
          if nt != t then
            raise
              (Failure (s ^ "has been bound to a different type previously"))
          else lb
      | None -> StringMap.add s t lb
  in
  let rec check_sutie (lb : typ StringMap.t) (lf : sfunc_decl StringMap.t)
      (ss : sutie) =
    match ss with
    | CaseSutie (s, e) ->
        let ((et, se) as e') = check_expr_logic lb lf e in
        if et != Bool then
          raise (Failure "the case is not a bool expression")
        else
          let (((lb, lf), (t, ss)) as s') = check_stmt lb lf s in
          (lb, (t, SCaseSutie (Some (t, ss), e')))
  and check_stmt (lb : typ StringMap.t) (lf : sfunc_decl StringMap.t)
      (ss : stmt) =
    match ss with
    | Expr e ->
        let ((t, se) as e') = check_expr_calc lb lf e in
        ((lb, lf), (t, SExpr e'))
    | Assign (s, e) ->
        let ((t, se) as e') = check_expr_calc lb lf e in
        ((add_bind (string_of_ident s) t lb lf, lf), (t, SAssign ((string_of_ident s), e')))
    | Print e ->
        let ((t, se) as e') = check_expr_calc lb lf e in
        ((lb, lf), (t, SPrint e'))
    | FuncDef (fname, formals, body) ->
        if StringMap.find_opt (string_of_ident fname) lb != None then
          raise (Failure ((string_of_ident fname) ^ "has been bound to a variable previously"))
        else if StringMap.find_opt (string_of_ident fname) lf != None then
          raise (Failure ((string_of_ident fname) ^ "has been bound to a function previously"))
        else
          let sd =
            {
              styp = Float;
              sfname = string_of_ident fname;
              sformals = List.map (fun a -> (Float, string_of_ident a)) formals;
              sframe = StringMap.fold (fun n t tl -> (t, n) :: tl) lb [];
              slocals = [];
              sbody = None;
            }
          in
          let llb =
            List.fold_left (fun bs a -> add_bind (string_of_ident a) Float bs lf) lb formals
          and llf = StringMap.add (string_of_ident fname) sd lf in
          let (slb, slf), (t, ssb) = check_stmt llb llf body in
          let not_exist k m = not (StringMap.mem k m) in
          let sl = StringMap.filter (fun k v -> not_exist k llb) slb in
          let sd =
            {
              styp = t;
              sfname = string_of_ident fname;
              sformals = List.map (fun a -> (Float, string_of_ident a)) formals;
              sframe = StringMap.fold (fun n t tl -> (t, n) :: tl) lb [];
              slocals = StringMap.fold (fun n t tl -> (t, n) :: tl) sl [];
              sbody = Some (t, ssb);
            }
          in
          ((lb, StringMap.add (string_of_ident fname) sd slf), (t, SFuncDef sd))
    | Case suites ->
        if
          List.exists
            (fun s -> fst (snd (check_sutie lb lf s)) = Float)
            suites
        then
          ( (lb, lf),
            ( Float,
              SCase (List.map (fun s -> snd (check_sutie lb lf s)) suites) )
          )
        else
          ( (lb, lf),
            ( Int,
              SCase (List.map (fun s -> snd (check_sutie lb lf s)) suites) )
          )
    | StmtClosure ss ->
        let (slb, slf), nss =
          List.fold_left_map (fun (b, f) s -> check_stmt b f s) (lb, lf) ss
        in
        ( (slb, slf),
          ( fst (List.hd (List.rev nss)),
            SStmtClosure (List.map (fun (t, s) -> Some (t, s)) nss) ) )
  in
  let (binds, func_decls), stmts =
    List.fold_left_map
      (fun (b, f) s -> check_stmt b f s)
      (StringMap.empty, StringMap.empty)
      prog
  in
  let maind =
    {
      styp = Int;
      sfname = "main";
      sformals = [];
      sframe = [];
      slocals = StringMap.fold (fun n t tl -> (t, n) :: tl) binds [];
      sbody =
        Some
          ( fst (List.hd (List.rev stmts)),
            SStmtClosure (List.map (fun (t, s) -> Some (t, s)) stmts) );
    }
  in
  let fds = StringMap.add "main" maind func_decls in
  List.rev (StringMap.fold (fun fn fd tl -> fd :: tl) fds [])
