
#!/bin/sh
SCANNER="./scanner"
MENHIR="menhir --interpret --interpret-show-cst ../../src/parser.mly"
REJECT_REF="test_cases/reject.reference"
# Set time limit for all operations
ulimit -t 30

globallog=test_parser.log
rm -f $globallog

tmp_output_dir=test_tmp
rm -rf $tmp_output_dir
mkdir $tmp_output_dir
error=0
globalerror=0

keep=0

Usage() {
    echo "Usage: test_parser.sh [options] [.ctex files]"
    echo "-k    Keep intermediate files"
    echo "-h    Print this help"
    exit 1
}

SignalError() {
    if [ $error -eq 0 ] ; then
	echo "FAILED"
	error=1
    fi
    echo "  $1"
}

# Compare <outfile> <reffile> <difffile>
# Compares the outfile with reffile.  Differences, if any, written to difffile
Compare() {
    generatedfiles="$generatedfiles $3"
    echo diff -b $1 $2 ">" $3 1>&2
    diff -b "$1" "$2" > "$3" 2>&1 || {
	SignalError "$1 differs"
	echo "FAILED $1 differs from $2" 1>&2
    }
}

# Run <args>
# Report the command, run it, and report any errors
Run() {
    echo $* 1>&2
    eval $* || {
	SignalError "$1 failed on $*"
	return 1
    }
}

# RunFail <args>
# Report the command, run it, and expect an error
RunFail() {
    echo $* 1>&2
    eval $* && {
	SignalError "failed: $* did not report an error"
	return 1
    }
    return 0
}

Check() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.ctex//'`
    reffile=`echo $1 | sed 's/.ctex$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."

    echo -n "$basename..."

    echo 1>&2
    echo "###### Testing $basename" 1>&2

    generatedfiles=""

    generatedfiles="$generatedfiles ${tmp_output_dir}/${basename}.out ${tmp_output_dir}/${basename}.diff" &&
    Run "$SCANNER" "<" "$1" "|" "$MENHIR" ">" "${tmp_output_dir}/${basename}.out" "2>>" $globallog &&
    Compare ${tmp_output_dir}/${basename}.out ${reffile}.reference ${tmp_output_dir}/${basename}.diff

    # Report the status and clean up the generated files

    if [ $error -eq 0 ] ; then
	if [ $keep -eq 0 ] ; then
	    rm -f $generatedfiles
	fi
	echo "OK"
	echo "###### SUCCESS" 1>&2
    else
	echo "###### FAILED" 1>&2
	globalerror=$error
    fi
}

CheckFail() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.ctex//'`
    reffile=`echo $1 | sed 's/.ctex$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."

    echo -n "$basename..."

    echo 1>&2
    echo "###### Testing $basename" 1>&2

    generatedfiles=""

    generatedfiles="$generatedfiles ${tmp_output_dir}/${basename}.err ${tmp_output_dir}/${basename}.diff" &&
    Run "$SCANNER" "<" "$1" "|" "$MENHIR" ">" "${tmp_output_dir}/${basename}.err" "2>>" $globallog &&
    Compare ${tmp_output_dir}/${basename}.err ${REJECT_REF} ${tmp_output_dir}/${basename}.diff

    # Report the status and clean up the generated files

    if [ $error -eq 0 ] ; then
	if [ $keep -eq 0 ] ; then
	    rm -f $generatedfiles
	fi
	echo "OK"
	echo "###### SUCCESS" 1>&2
    else
	echo "###### FAILED" 1>&2
	globalerror=$error
    fi
}

while getopts kdpsh c; do
    case $c in
	k) # Keep intermediate files
	    keep=1
	    ;;
	h) # Help
	    Usage
	    ;;
    esac
done


if [ $# -ge 1 ]
then
    files=$@
else
    files="test_cases/test_*.ctex test_cases/bad_*.ctex"
fi

for file in $files
do
    case $file in
    *bad_*)
	    CheckFail $file 2>> $globallog
	    ;;
	*test_*)
	    Check $file 2>> $globallog
	    ;;
	*)
	    echo "unknown file type $file"
	    globalerror=1
	    ;;
    esac
done

exit $globalerror
